from syntaxtrees.ParseNode import *
from syntaxtrees.AST import *


globalvars = {}       # We will store the calculator's variables here

def lookup(map, name):
    for x,v in map:
        if x==name: return v
    if name not in globalvars.keys(): print 'Undefined:', name
    return globalvars.get(name, 0)

%%
parser FormulaParser:
    ignore:    "[ \r\t\n]+"

    token int: "[0-9]+"
    token real: "([0-9]+[.][0-9]*)|([0-9]*[.][0-9]+)"
    token number: "([0-9]+)|(([0-9]+[.][0-9]*)|([0-9]*[.][0-9]+))"
    token char: "[a-zA-Z]"
    token string: "[0-9]*[a-zA-Z]+([a-zA-Z]|[0-9])*"
    token ref: "[$]?[A-Z][A-Z]?[A-Z]?[$]?[1-9][0-9]?[0-9]?[0-9]?"
    token rangeref: "[$]?[A-Z][A-Z]?[A-Z]?[$]?[1-9][0-9]?[0-9]?[0-9]?[:][$]?[A-Z][A-Z]?[A-Z]?[$]?[1-9][0-9]?[0-9]?[0-9]?"
    token addop: "[+-]"
    token mulop: "[*/]"
    token add: "\+"
    token sub: "-"
    token mul: "\*"
    token div: "/"
    token expop: "\^"
    token funcname: "[a-zA-Z]+"
    token lpar: "\("
    token rpar: "\)"
    token comma: ","
    token equal: "="
    token empty: ""


    token end: ";"
    token NUM: "[0-9]+"
    token VAR: "[a-zA-Z_]+"


    rule GOAL:          (FORMULA                        {{return FORMULA}}

                      | string) end                     {{return ParseNode("string", string)}}

    rule EXP:            TERM                           {{parseNode=ParseNode("exp")}}
                                                        {{parseNode.children.append(TERM)}}

                        ((add TERM)                     {{parseNode.children.append(ParseNode("add"))}}
                                                        {{parseNode.children.append(TERM)}}
                      |  (sub TERM)                     {{parseNode.children.append(ParseNode("sub"))}}
                                                        {{parseNode.children.append(TERM)}}
                        )*                              {{return parseNode}}

    rule TERM:          EXPTERM                         {{parseNode=ParseNode("term")}}
                                                        {{parseNode.children.append(EXPTERM)}}
                        ((mul EXPTERM)                  {{parseNode.children.append(ParseNode("mul"))}}
                                                        {{parseNode.children.append(EXPTERM)}}
                      |  (div EXPTERM)                  {{parseNode.children.append(ParseNode("div"))}}
                                                        {{parseNode.children.append(EXPTERM)}}
                        )*                              {{return parseNode}}

    rule EXPTERM:       SIGNEDFACTOR                    {{parseNode=ParseNode("expterm")}}
                                                        {{parseNode.children.append(SIGNEDFACTOR)}}
                        (expop SIGNEDFACTOR             {{parseNode.children.append(ParseNode("exponent"))}}
                                                        {{parseNode.children.append(SIGNEDFACTOR)}}
                        )*                              {{return parseNode}}

    rule SIGNEDFACTOR:   FACTOR                         {{return FACTOR}}
                      |
                        (sub FACTOR)                    {{parseNode=ParseNode("signedfactor")}}
                                                        {{parseNode.children.append(FACTOR)}}
                                                        {{return parseNode}}

    rule FACTOR:        lpar EXP rpar                   {{return EXP}}
                      | int                             {{return ParseNode("number",int)}}
                      | real                            {{return ParseNode("number",real)}}
                      | ref                             {{return ParseNode("ref", ref)}}
                      | FUNCTION                        {{return FUNCTION}}


    rule ARG:                                           {{parseNode=ParseNode("arg")}}

                      (rangeref                         {{parseNode.children.append(ParseNode("rangeref", rangeref))}}
                    |  EXP)                             {{parseNode.children.append(EXP)}}
                        (comma
                          (rangeref                     {{parseNode.children.append(ParseNode("rangeref", rangeref))}}
                         | EXP)                         {{parseNode.children.append(EXP)}}
                        )*                              {{return parseNode}}

    rule FUNCTION:      funcname                        {{parseNode=ParseNode("function", funcname)}}
                                                        {{ARG=None}}
                        lpar
                        [ARG]                           {{if ARG!=None:parseNode.children.append(ARG)}}
                        rpar
                                                        {{return parseNode}}

    rule FORMULA:       '='EXP                          {{return EXP}}


%%
def getCellRef(ref):
    absoluteColumn=0
    absoluteRow=0
    column=""
    row=""
    i=0
    #Checking if the column is absolute
    if ref[i]=='$':
        absoluteColumn=1
        i+=1
    while not ((ref[i]>='0' and ref[i]<='9') or ref[i]=='$'):
        column+=ref[i]
        i+=1

    column=stringToInt(column)

    #Checking if the row is absolute
    if ref[i]=='$':
        absoluteRow=1
        i+=1
    while i<len(ref):
        row+=ref[i]
        i+=1
    row=int(row)

    return CellRef(column, row, absoluteColumn,absoluteRow)

def constructAST(parseTree):
    if parseTree.kind=="number":
        return Number(float(parseTree.strData))
    elif parseTree.kind=="ref":
        return getCellRef(parseTree.strData)
    elif parseTree.kind=="rangeref":
        refs=parseTree.strData.split(":")
        return RangeRef(getCellRef(refs[0]),
                        getCellRef(refs[1]))
    elif parseTree.kind=="function":
        args=[]
        if len(parseTree.children)>0:
            args=constructAST(parseTree.children[0])
        return Function(parseTree.strData, args)
    elif parseTree.kind=="arg":
        args=[]
        for arg in parseTree.children:
            args.append(constructAST(arg))
        return args
    elif parseTree.kind=="exp":
        i=0
        if len(parseTree.children)==1:
            return constructAST(parseTree.children[0])
        else:
            operands=[]
            while i<len(parseTree.children):
                if i==0:
                    operands.append(constructAST(parseTree.children[i]))
                    i+=2
                elif parseTree.children[i-1].kind=="add":
                    operands.append(constructAST(parseTree.children[i]))
                    i+=2
                else:
                    operands.append(Function("invSum", [constructAST(parseTree.children[i])]))
                    i+=2
            return Function("sum", operands)
    elif parseTree.kind=="term":
        i=0
        if len(parseTree.children)==1:
            return constructAST(parseTree.children[0])
        else:
            operands=[]
            while i<len(parseTree.children):
                if i==0:
                    operands.append(constructAST(parseTree.children[i]))
                    i+=2
                elif parseTree.children[i-1].kind=="mul":
                    operands.append(constructAST(parseTree.children[i]))
                    i+=2
                else:
                    operands.append(Function("invMul", [constructAST(parseTree.children[i])]))
                    i+=2
            return Function("mul", operands)
    elif parseTree.kind=="expterm":
        i=0
        if len(parseTree.children)==1:
            return constructAST(parseTree.children[0])
        else:
            operands=[]
            while i<len(parseTree.children):
                operands.append(constructAST(parseTree.children[i]))
                i+=2
            return Function("exp", operands)
    elif parseTree.kind=="signedfactor":
        return Function("invSum", [constructAST(parseTree.children[0])])
    else:
        raise ValueError("Wrong type for simple expression " + parseTree.kind)


def getAST(stringToParse):

    parseTree=parse("GOAL", stringToParse)

    # parseTree returned by parse() is a structure (A,B,C)
    #
    # A: if 1, parsing success; if 0, parsing failure
    # B: contains the parse tree
    # C: contains an informative string 
    #    ('' when success, otherwise errormsg)

    if parseTree[0]:
        return (1,constructAST(parseTree[1]),parseTree[2])
    else:
        return (0,None,parseTree[2])

